pub mod additional_rfactor;
#[allow(warnings)]
mod rfactor_structs;

use std::{cell::Cell, fs::File, marker::PhantomData, path::Path};

use additional_rfactor::*;
pub use additional_rfactor::{rF2Vec3, rF2VehicleTelemetry, rF2Wheel};
use anyhow::Result;
use memmap2::Mmap;
pub use rfactor_structs::{ScoringInfoV01, VehicleScoringInfoV01};

const RFACTOR_SHM_FILE: &str = "/dev/shm";

struct ShMReader<T> {
    _file: File,
    shm: Mmap,

    version: Cell<VersionHeader>,

    data: PhantomData<T>,
}

impl<T> ShMReader<T> {
    const SIZE: usize = std::mem::size_of::<T>();

    fn new(mm_file_name: &str) -> Result<Self> {
        let file = File::open(Path::new(RFACTOR_SHM_FILE).join(mm_file_name))?;
        let mmap = unsafe { Mmap::map(&file)? };

        Ok(Self {
            _file: file,
            shm: mmap,

            version: Cell::new(VersionHeader::default()),

            data: PhantomData,
        })
    }

    fn check_version_update(&self) -> bool {
        let new_version = VersionHeader::from(&self.shm[0..VersionHeader::SIZE]);

        let are_different = self.version.get().check_for_update(&new_version);

        if are_different {
            self.version.set(new_version);
        }

        are_different
    }
}

impl<'a, T> ShMReader<T>
where
    T: From<&'a [u8]>,
{
    fn read(&'a self) -> Option<T> {
        if self.check_version_update() {
            Some(T::from(&self.shm[0..Self::SIZE]))
        } else {
            None
        }
    }
}

pub struct TelemetryReader {
    mm_reader: ShMReader<rF2Telemetry>,
}

impl TelemetryReader {
    pub fn new() -> Result<Self> {
        Ok(Self {
            mm_reader: ShMReader::new(MM_TELEMETRY_FILE_NAME)?,
        })
    }

    pub fn query_telemetry(&self) -> Option<Vec<rF2VehicleTelemetry>> {
        self.mm_reader.read().map(|v| v.vehicles().to_vec())
    }
}

pub struct ScoringReader {
    mm_reader: ShMReader<rF2Scoring>,
}

impl ScoringReader {
    pub fn new() -> Result<Self> {
        Ok(Self {
            mm_reader: ShMReader::new(MM_SCORING_FILE_NAME)?,
        })
    }

    pub fn vehicle_scoring(&self) -> Option<(ScoringInfoV01, Vec<VehicleScoringInfoV01>)> {
        self.mm_reader
            .read()
            .map(|s| (s.scoring_info, s.vehicles().to_vec()))
    }
}
